/*
 * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
 * Copyright(c) 1995-99 Andrew Lister
 * Copyright  1999, 2000, 2001, 2002, 2003 by the LessTif Developers.
 *
 *                        All rights reserved
 * Permission to use, copy, modify and distribute this material for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Bellcore not be used in advertising
 * or publicity pertaining to this material without the specific,
 * prior written permission of an authorized representative of
 * Bellcore.
 *
 * BELLCORE MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES, EX-
 * PRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR ANY PARTICULAR PURPOSE, AND THE WARRANTY AGAINST IN-
 * FRINGEMENT OF PATENTS OR OTHER INTELLECTUAL PROPERTY RIGHTS.  THE
 * SOFTWARE IS PROVIDED "AS IS", AND IN NO EVENT SHALL BELLCORE OR
 * ANY OF ITS AFFILIATES BE LIABLE FOR ANY DAMAGES, INCLUDING ANY
 * LOST PROFITS OR OTHER INCIDENTAL OR CONSEQUENTIAL DAMAGES RELAT-
 * ING TO THE SOFTWARE.
 *
 * $Id: Actions.c,v 1.56 2004/08/04 06:11:14 dannybackx Exp $
 */

#define	VERBOSE_SLIDE

/*
 * Actions.c created by Andrew Lister (7 August, 1995)
 */

#ifdef HAVE_CONFIG_H
#include <XbaeConfig.h>
#endif

#include <stdio.h>

#include <Xm/Xm.h>
#include <Xm/XmP.h>
#include <Xm/DrawP.h>
#include <Xm/ScrollBar.h>
#include <Xbae/MatrixP.h>
#include <Xbae/Clip.h>
#include <Xbae/Draw.h>
#include <Xbae/Actions.h>
#include <Xbae/Utils.h>
#include <Xbae/ScrollMgr.h>
#include <Xbae/ClipP.h>
#include <X11/cursorfont.h>

#include <XbaeDebug.h>


#ifndef XlibSpecificationRelease
#define XrmPermStringToQuark XrmStringToQuark
#endif

#if !defined(DRAW_RESIZE_LINE) && !defined(DRAW_RESIZE_SHADOW)

/* One of DRAW_RESIZE_LINE and DRAW_RESIZE_SHADOW must be defined. */
#define DRAW_RESIZE_SHADOW
#endif

#ifndef DEFAULT_SCROLL_SPEED
#define DEFAULT_SCROLL_SPEED 150
#endif

/*
 * This is the number of pixels that you can 'miss' the cell edge when
 * trying to resize a row or column.
 */
#ifndef XBAE_RESIZE_FUZZ
#define XBAE_RESIZE_FUZZ	4
#endif

typedef struct {
        XbaeMatrixWidget mw;
        GC gc;
        int row, column;
        int startx, starty;
        int lastx, lasty;
        int currentx, currenty;
        int x, width, y, height;
        short *columnWidths, *rowHeights;
        Boolean grabbed;
        Boolean haveVSB, haveHSB;
} XbaeMatrixResizeColumnStruct;

typedef struct {
        XbaeMatrixWidget mw;
        int row;
        int column;
        Boolean pressed;
        Boolean grabbed;
} XbaeMatrixButtonPressedStruct;

typedef struct {
        XbaeMatrixWidget mw;
        XEvent *event;
        XtIntervalId timerID;
        XtAppContext app_context;
        unsigned long interval;
        Boolean grabbed;
        int currentx;
        int currenty;
        int repeat_count;
        int scroll_region;
} XbaeMatrixScrollStruct;

static int DoubleClick(XbaeMatrixWidget, XEvent *, int, int);
static void DrawSlideRow(XbaeMatrixWidget, int);
static void DrawSlideColumn(XbaeMatrixWidget, int);
static void SlideColumn(Widget, XtPointer, XEvent *, Boolean *);
static void PushButton(Widget, XtPointer, XEvent *, Boolean *);
static void updateScroll(XtPointer);
static void checkScrollValues(Widget, XtPointer, XEvent *, Boolean *);
static void callSelectCellCallbacks(XbaeMatrixWidget, XEvent *, int, int, String *, Cardinal);

static Boolean did_move = False;
static int last_row = -1;
static int last_column = -1;

/* ARGSUSED */
void xbaeDefaultActionACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        int x, y;
        int row, column;

        DEBUGOUT(_XbaeDebug
                 (__FILE__, w, "xbaeDefaultActionACT(x %d y %d)\n", event->xbutton.x,
                  event->xbutton.y));

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */

        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "defaultActionACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to DefaultAction action", NULL, 0);
                return;
        }

        if (!mw->matrix.default_action_callback)
                return;

        if (!xbaeEventToMatrixXY(mw, event, &x, &y))
                return;

        if (!xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column))
                return;

        XmProcessTraversal((Widget) mw, XmTRAVERSE_CURRENT);

        if (DoubleClick(mw, event, row, column)) {
                XbaeMatrixDefaultActionCallbackStruct call_data;

                call_data.reason = XbaeDefaultActionReason;
                call_data.event = event;
                call_data.row = row;
                call_data.column = column;

                XtCallCallbackList((Widget) mw, mw->matrix.default_action_callback,
                                   (XtPointer) & call_data);
        }
}




/* ARGSUSED */
void xbaeHandleTrackingACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixTrackCellCallbackStruct call_data;

        XbaeMatrixWidget mw;
        int x, y;
        int row, column;

        DEBUGOUT(_XbaeDebug
                 (__FILE__, w, "%s(x %d y %d)\n", __FUNCTION__, event->xbutton.x,
                  event->xbutton.y));

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */

        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "defaultActionACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to DefaultAction action", NULL, 0);
                return;
        }

        if (!mw->matrix.default_action_callback)
                return;


        if (!xbaeEventToMatrixXY(mw, event, &x, &y))
                return;

        call_data.pointer_x = x;
        call_data.pointer_y = y;

        (void) xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column);

        if ((column != mw->matrix.prev_column) || (row != mw->matrix.prev_row)) {
                call_data.reason = XbaeTrackCellReason;
                call_data.event = event;
                call_data.prev_row = mw->matrix.prev_row;
                call_data.prev_column = mw->matrix.prev_column;
                call_data.row = row;
                call_data.column = column;

                /* printf ("%s %s %d  XY:%d %d RC: %d %d\n",
                 * __FILE__,__FUNCTION__,__LINE__,x, y, row, column);
                 */

                XtCallCallbackList((Widget) mw, mw->matrix.track_cell_callback,
                                   (XtPointer) & call_data);

                /*
                 * Merken
                 */
                mw->matrix.prev_column = column;
                mw->matrix.prev_row = row;
        }
}

static void DrawSlideColumn(XbaeMatrixWidget mw, int x)
{
#ifdef DRAW_RESIZE_SHADOW
        /* These values derived through that age-old process
         * of what looks good to me */
#define SHADOW_WIDTH 2
#define RESIZE_COLUMN_LINE_WIDTH 4
        Dimension width = RESIZE_COLUMN_LINE_WIDTH;
        Dimension shadow_width = SHADOW_WIDTH;
#endif
        Dimension height;
        Window win;
        Display *display = XtDisplay(mw);
        int column = xbaeXtoCol(mw, x - COLUMN_LABEL_OFFSET(mw));
        int top, bottom;
        int adjusted_x;
        int y;
#ifdef DRAW_RESIZE_LINE
        GC gc = mw->matrix.draw_gc;
#endif
        Boolean need_vert_dead_space_fill = NEED_VERT_DEAD_SPACE_FILL(mw);
        unsigned int clip_reason;

        /*
         * If the column being resized is a fixed one then we don't need to
         * bother with the clip region
         */
        if (column < (int) mw->matrix.fixed_columns) {
                y = ROW_LABEL_OFFSET(mw);
                height = VISIBLE_HEIGHT(mw) + FIXED_ROW_HEIGHT(mw) + TRAILING_FIXED_ROW_HEIGHT(mw);
                win = XtWindow(mw);

                if (need_vert_dead_space_fill)
                        height += VERT_DEAD_SPACE_HEIGHT(mw);

#ifdef DRAW_RESIZE_LINE
                XDrawLine(display, win, gc, x, y, x, y + height);
                if (XtIsManaged(LeftClip(mw)))
                        XDrawLine(display, XtWindow(LeftClip(mw)), gc, x - COLUMN_LABEL_OFFSET(mw),
                                  0, x - COLUMN_LABEL_OFFSET(mw), LeftClip(mw)->core.height);
#endif
#ifdef DRAW_RESIZE_SHADOW
                DRAW_SHADOW(display, win, "win 1", mw->matrix.resize_top_shadow_gc,
                            mw->matrix.resize_bottom_shadow_gc, shadow_width, x, y, width, height,
                            XmSHADOW_OUT);
                if (XtIsManaged(LeftClip(mw)))
                        DRAW_SHADOW(display, XtWindow(LeftClip(mw)), "LeftClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width,
                                    x - COLUMN_LABEL_OFFSET(mw), 0, width,
                                    LeftClip(mw)->core.height, XmSHADOW_OUT);
#endif
                return;
        }

        /*
         * Similarly for trailingFixedColumns - beware going off the clip child
         * here also
         */
        if (column >= TRAILING_HORIZ_ORIGIN(mw)
            || x >= (int) (ClipChild(mw)->core.x + ClipChild(mw)->core.width)) {
                y = ROW_LABEL_OFFSET(mw);
                height = VISIBLE_HEIGHT(mw) + FIXED_ROW_HEIGHT(mw) + TRAILING_FIXED_ROW_HEIGHT(mw);
                win = XtWindow(mw);

                if (need_vert_dead_space_fill)
                        height += VERT_DEAD_SPACE_HEIGHT(mw);

#ifdef DRAW_RESIZE_LINE
                XDrawLine(display, win, gc, x, y, x, y + height);
                if (XtIsManaged(RightClip(mw)))
                        XDrawLine(display, XtWindow(RightClip(mw)), gc,
                                  x - TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw), 0,
                                  x - TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw),
                                  RightClip(mw)->core.height);

#endif
#ifdef DRAW_RESIZE_SHADOW
                DRAW_SHADOW(display, win, "win 2", mw->matrix.resize_top_shadow_gc,
                            mw->matrix.resize_bottom_shadow_gc, shadow_width, x, y, width, height,
                            XmSHADOW_OUT);
                if (XtIsManaged(RightClip(mw)))
                        DRAW_SHADOW(display, XtWindow(RightClip(mw)), "RightClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width,
                                    x - TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw), 0, width,
                                    RightClip(mw)->core.height, XmSHADOW_OUT);
#endif
                return;
        }

        xbaeGetVisibleRows(mw, &top, &bottom);
        /*
         * we need all non-fixed rows, so add 1 to bottom
         * to include the last one as the return values
         * are inclusive
         */
        bottom += 1;

        /*
         * The area between top and bottom rows are the non fixed rows.  They
         * fall on the ClipChild
         */
        y = -mw->matrix.cell_shadow_thickness;  /* relative to clip */

        height =
            ROW_POSITION(mw, bottom) - ROW_POSITION(mw, top) + 2 * mw->matrix.cell_shadow_thickness;

        /*
         * If we are on the clip, the x location is offset by the
         * fixed column width, label offset and label width
         */
        adjusted_x = x - FIXED_COLUMN_LABEL_OFFSET(mw);

        win = XtWindow(ClipChild(mw));

#ifdef DRAW_RESIZE_LINE
        XDrawLine(display, win, gc, adjusted_x, y, adjusted_x, y + height);
#endif
#ifdef DRAW_RESIZE_SHADOW
        DRAW_SHADOW(display, win, "win 3", mw->matrix.resize_top_shadow_gc,
                    mw->matrix.resize_bottom_shadow_gc, shadow_width, adjusted_x, y, width, height,
                    XmSHADOW_OUT);
#endif
        /*
         * Now draw the line (or shadow) on the non clipped region - that is
         * the fixed and trailingFixed rows.  First, do the leading rows.
         */
        if (mw->matrix.fixed_rows) {
                y = ROW_LABEL_OFFSET(mw);
                height = FIXED_ROW_HEIGHT(mw) + 2 * mw->matrix.cell_shadow_thickness;
                win = XtWindow(mw);
                xbaeSetClipMask(mw, CLIP_FIXED_ROWS);

#ifdef DRAW_RESIZE_LINE
                if (XtIsManaged(TopClip(mw)))
                        XDrawLine(display, XtWindow(TopClip(mw)), gc, adjusted_x,
                                  -mw->matrix.cell_shadow_thickness, adjusted_x, height);
#endif
#ifdef DRAW_RESIZE_SHADOW
                if (XtIsManaged(TopClip(mw)))
                        DRAW_SHADOW(display, XtWindow(TopClip(mw)), "TopClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width, adjusted_x,
                                    -mw->matrix.cell_shadow_thickness, width, height, XmSHADOW_OUT);
#endif
                xbaeSetClipMask(mw, CLIP_NONE);
        }

        /*
         * The trailingFixedRows
         */
        if (mw->matrix.trailing_fixed_rows) {
                y = TRAILING_FIXED_ROW_LABEL_OFFSET(mw);
                height = TRAILING_FIXED_ROW_HEIGHT(mw) + 2 * mw->matrix.cell_shadow_thickness;

                clip_reason = CLIP_TRAILING_FIXED_ROWS;
                if (IS_LEADING_FIXED_COLUMN(mw, column))
                        clip_reason |= CLIP_FIXED_COLUMNS;
                else if (IS_TRAILING_FIXED_COLUMN(mw, column))
                        clip_reason |= CLIP_TRAILING_FIXED_COLUMNS;

                xbaeSetClipMask(mw, clip_reason);

#ifdef DRAW_RESIZE_LINE
                if (XtIsManaged(BottomClip(mw)))
                        XDrawLine(display, XtWindow(BottomClip(mw)), gc, adjusted_x,
                                  -mw->matrix.cell_shadow_thickness, adjusted_x, height);
#endif
#ifdef DRAW_RESIZE_SHADOW
                if (XtIsManaged(BottomClip(mw)))
                        DRAW_SHADOW(display, XtWindow(BottomClip(mw)), "BottomClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width, adjusted_x,
                                    -mw->matrix.cell_shadow_thickness, width, height, XmSHADOW_OUT);
#endif
                xbaeSetClipMask(mw, CLIP_NONE);
        }

        if ((NEED_VERT_FILL(mw) && (!HAS_ATTACHED_TRAILING_ROWS(mw))) || need_vert_dead_space_fill) {
                if (need_vert_dead_space_fill) {
                        y = UNATTACHED_TRAILING_ROWS_OFFSET(mw) - mw->matrix.cell_shadow_thickness;
                        height = 2 * mw->matrix.cell_shadow_thickness + VERT_DEAD_SPACE_HEIGHT(mw);
                } else {
                        y = TRAILING_FIXED_ROW_LABEL_OFFSET(mw) + TRAILING_FIXED_ROW_HEIGHT(mw);
                        height = FILL_VERT_HEIGHT(mw);
                }

#ifdef DRAW_RESIZE_LINE
                XDrawLine(display, XtWindow(mw), gc, adjusted_x, y, adjusted_x, height);
#endif
#ifdef DRAW_RESIZE_SHADOW
                DRAW_SHADOW(display, XtWindow(mw), "win 4", mw->matrix.resize_top_shadow_gc,
                            mw->matrix.resize_bottom_shadow_gc, shadow_width, x, y, width, height,
                            XmSHADOW_OUT);
#endif
        }
}

static void SlideColumn(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
        XbaeMatrixResizeColumnStruct *rd = (XbaeMatrixResizeColumnStruct *) data;
        XMotionEvent *motionEvent;
        Boolean relayout = False;
        int numCharacters;
        int i;
        int HORIZ_RESIZE_WIDTH = 1;

        if (!rd->mw->matrix.column_width_in_pixels)
                HORIZ_RESIZE_WIDTH = FONT_WIDTH(rd->mw);


        if (event->type == ButtonRelease) {
                DrawSlideColumn(rd->mw, rd->lastx);
                XUngrabPointer(XtDisplay(w), CurrentTime);
                rd->grabbed = False;
                /*
                 * Remanage the VSB if we unmapped it earlier
                 */
                if (rd->haveVSB)
                        XtManageChild(VertScrollChild(rd->mw));

                if (rd->mw->matrix.resize_column_callback) {
                        XbaeMatrixResizeColumnCallbackStruct call_data;

                        call_data.reason = XbaeResizeColumnReason;
                        call_data.event = event;
                        call_data.row = rd->row;
                        call_data.column = rd->column - 1;
                        call_data.which = rd->column - 1;
                        call_data.columns = rd->mw->matrix.columns;
                        call_data.column_widths = rd->columnWidths;
                        XtCallCallbackList((Widget) rd->mw, rd->mw->matrix.resize_column_callback,
                                           (XtPointer) & call_data);
                }

                for (i = 0; i < rd->mw->matrix.columns; i++)
                        if (rd->columnWidths[i] != rd->mw->matrix.column_widths[i]) {
                                /* Make sure everything is handled correctly with SetValues */
                                XtVaSetValues((Widget) rd->mw, XmNcolumnWidths, rd->columnWidths,
                                              NULL);
                                break;
                        }
                /*
                 * If maxColumnLengths are set and we have resized the column to
                 * larger, reset the corresponding maxColumnLength
                 */
                if (rd->mw->matrix.column_max_lengths
                    && rd->columnWidths[rd->column - 1] >
                    rd->mw->matrix.column_max_lengths[rd->column - 1])
                        rd->mw->matrix.column_max_lengths[rd->column - 1] =
                            rd->columnWidths[rd->column - 1];
                XtFree((char *) rd->columnWidths);
                return;
        }

        if (event->type != MotionNotify)        /* Double check! */
                return;

        motionEvent = (XMotionEvent *) event;

        if (rd->currentx - motionEvent->x > HORIZ_RESIZE_WIDTH) {
                /* If we're only one character wide, we cannae get any smaller */
                if (rd->columnWidths[rd->column - 1] == BAD_WIDTH + 1)
                        return;
                /*
                 * Moved left a full character - update the column widths and force
                 * a redisplay
                 */
                if (rd->mw->matrix.column_width_in_pixels) {
                        numCharacters = (rd->currentx - motionEvent->x);
                        if (numCharacters >= rd->columnWidths[rd->column - 1])
                                /* Must keep a column at least one character wide */
                                numCharacters = rd->columnWidths[rd->column - 1] - 1;
                        rd->columnWidths[rd->column - 1] -= numCharacters;
                        rd->currentx -= numCharacters;
                } else {
                        numCharacters = (rd->currentx - motionEvent->x) / FONT_WIDTH(rd->mw);
                        if (numCharacters >= rd->columnWidths[rd->column - 1])
                                /* Must keep a column at least one character wide */
                                numCharacters = rd->columnWidths[rd->column - 1] - 1;

                        rd->columnWidths[rd->column - 1] -= numCharacters;
                        rd->currentx -= numCharacters * FONT_WIDTH(rd->mw);
                }
                relayout = True;
        }

        if (motionEvent->x - rd->currentx > HORIZ_RESIZE_WIDTH) {
                /*
                 * Moved right a full character - update the column widths and force
                 * a redisplay
                 */
                if (rd->mw->matrix.column_width_in_pixels) {
                        numCharacters = (motionEvent->x - rd->currentx);
                        rd->columnWidths[rd->column - 1] += numCharacters;
                        rd->currentx += numCharacters;
                } else {
                        numCharacters = (motionEvent->x - rd->currentx) / FONT_WIDTH(rd->mw);
                        rd->columnWidths[rd->column - 1] += numCharacters;
                        rd->currentx += numCharacters * FONT_WIDTH(rd->mw);
                }
                relayout = True;
        }

        if (relayout) {
                /* Draw the marker line in the new location */
                if (rd->lastx != rd->currentx) {
                        DrawSlideColumn(rd->mw, rd->currentx);
                        DrawSlideColumn(rd->mw, rd->lastx);

                        rd->lastx = rd->currentx;
                }
        }
}

/* Derived from DrawSlideColumn */
static void DrawSlideRow(XbaeMatrixWidget mw, int y)
{
#ifdef DRAW_RESIZE_SHADOW
        /* These values derived through that age-old process
         * of what looks good to me */
#define SHADOW_WIDTH 2
#define RESIZE_ROW_LINE_WIDTH 4
        Dimension height = RESIZE_ROW_LINE_WIDTH;
        Dimension shadow_width = SHADOW_WIDTH;
#endif
        Dimension width;
        Window win;
        Display *display = XtDisplay(mw);
        int row = xbaeYtoRow(mw, y - ROW_LABEL_OFFSET(mw));
        int left, right;
        int adjusted_y;
        int x;
#ifdef DRAW_RESIZE_LINE
        GC gc = mw->matrix.draw_gc;
#endif
        Boolean need_horiz_dead_space_fill = NEED_HORIZ_DEAD_SPACE_FILL(mw);
        unsigned int clip_reason;

#ifdef	VERBOSE_SLIDE
        DEBUGOUT(_XbaeDebug(__FILE__, (Widget) mw, "DrawSlideRow(%d)\n", y));
#endif

        /*
         * If the row being resized is a fixed one then we don't need to
         * bother with the clip region
         */
        if (row < (int) mw->matrix.fixed_rows) {
                x = COLUMN_LABEL_OFFSET(mw);
                width =
                    VISIBLE_WIDTH(mw) + FIXED_COLUMN_WIDTH(mw) + TRAILING_FIXED_COLUMN_WIDTH(mw);
                win = XtWindow(mw);

                if (need_horiz_dead_space_fill)
                        width += HORIZ_DEAD_SPACE_WIDTH(mw);

#ifdef DRAW_RESIZE_LINE
#ifdef	VERBOSE_SLIDE
                DEBUGOUT(_XbaeDebug
                         (__FILE__, (Widget) mw, "XDrawLine(%d,%d,%d,%d)\n", x, y, x + width, y));
#endif
                XDrawLine(display, win, gc, x, y, x + width, y);
                if (XtIsManaged(TopClip(mw))) {
#ifdef	VERBOSE_SLIDE
                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw, "XDrawLine(%d,%d,%d,%d)\n", 0,
                                  y - ROW_LABEL_OFFSET(mw), TopClip(mw)->core.width,
                                  y - ROW_LABEL_OFFSET(mw)));
#endif
                        XDrawLine(display, XtWindow(TopClip(mw)), gc, 0, y - ROW_LABEL_OFFSET(mw),
                                  TopClip(mw)->core.width, y - ROW_LABEL_OFFSET(mw));
                }
#endif
#ifdef DRAW_RESIZE_SHADOW
                DRAW_SHADOW(display, win, "win 5", mw->matrix.resize_top_shadow_gc,
                            mw->matrix.resize_bottom_shadow_gc, shadow_width, x, y, width, height,
                            XmSHADOW_OUT);
                if (XtIsManaged(TopClip(mw)))
                        DRAW_SHADOW(display, XtWindow(TopClip(mw)), "TopClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width, 0,
                                    y - ROW_LABEL_OFFSET(mw), TopClip(mw)->core.width, height,
                                    XmSHADOW_OUT);
#endif
                return;
        }
        /*
         * Similarly for trailingFixedRows - beware going off the clip child
         * here also
         */
        if (row >= TRAILING_VERT_ORIGIN(mw)
            || y >= (int) (ClipChild(mw)->core.y + ClipChild(mw)->core.height)) {
                x = COLUMN_LABEL_OFFSET(mw);
                width =
                    VISIBLE_WIDTH(mw) + FIXED_COLUMN_WIDTH(mw) + TRAILING_FIXED_COLUMN_WIDTH(mw);
                win = XtWindow(mw);

                if (need_horiz_dead_space_fill)
                        width += HORIZ_DEAD_SPACE_WIDTH(mw);

#ifdef DRAW_RESIZE_LINE
#ifdef	VERBOSE_SLIDE
                DEBUGOUT(_XbaeDebug
                         (__FILE__, (Widget) mw, "XDrawLine(%d,%d,%d,%d)\n", x, y, x + width, y));
#endif
                XDrawLine(display, win, gc, x, y, x + width, y);
                if (XtIsManaged(BottomClip(mw))) {      /* Used to be RightClip */
#ifdef	VERBOSE_SLIDE
                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw, "XDrawLine(BottomClip %d,%d,%d,%d)\n", 0,
                                  y - TRAILING_FIXED_ROW_LABEL_OFFSET(mw),
                                  BottomClip(mw)->core.width,
                                  y - TRAILING_FIXED_ROW_LABEL_OFFSET(mw)));
#endif
                        XDrawLine(display, XtWindow(BottomClip(mw)), gc, 0,
                                  y - TRAILING_FIXED_ROW_LABEL_OFFSET(mw),
                                  BottomClip(mw)->core.width,
                                  y - TRAILING_FIXED_ROW_LABEL_OFFSET(mw));
                }
#endif
#ifdef DRAW_RESIZE_SHADOW
                DRAW_SHADOW(display, win, "win 6", mw->matrix.resize_top_shadow_gc,
                            mw->matrix.resize_bottom_shadow_gc, shadow_width, x, y, width, height,
                            XmSHADOW_OUT);
                if (XtIsManaged(BottomClip(mw)))
                        DRAW_SHADOW(display, XtWindow(BottomClip(mw)), "BottopClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width, 0,
                                    y - TRAILING_FIXED_ROW_LABEL_OFFSET(mw),
                                    BottomClip(mw)->core.width, height, XmSHADOW_OUT);
#endif
                return;
        }

        xbaeGetVisibleColumns(mw, &left, &right);
        /*
         * we need all non-fixed columns, so add 1 to left to include the last one
         * as the return values are inclusive
         left += 1;
         */
        right += 1;

        /*
         * The area between left and right rows are the non fixed columns.
         * They fall on the ClipChild
         */
        x = -mw->matrix.cell_shadow_thickness;  /* relative to clip */

        width = VISIBLE_WIDTH(mw) + FIXED_COLUMN_WIDTH(mw) + TRAILING_FIXED_COLUMN_WIDTH(mw);

#ifdef	VERBOSE_SLIDE
        DEBUGOUT(_XbaeDebug
                 (__FILE__, (Widget) mw, "Width %d = %d - %d + 2 * %d (left %d, right %d)\n", width,
                  COLUMN_POSITION(mw, right), COLUMN_POSITION(mw, left),
                  mw->matrix.cell_shadow_thickness, left, right));
#endif

        /*
         * If we are on the clip, the y location is offset by the
         * fixed row height, label offset and label height
         */
        adjusted_y = y - FIXED_ROW_LABEL_OFFSET(mw);

#ifdef	VERBOSE_SLIDE
        DEBUGOUT(_XbaeDebug(__FILE__, (Widget) mw, "DrawSlideRow(%d) adjust %d\n", y, adjusted_y));
#endif

        win = XtWindow(ClipChild(mw));

#ifdef DRAW_RESIZE_LINE
#ifdef	VERBOSE_SLIDE
        DEBUGOUT(_XbaeDebug
                 (__FILE__, (Widget) mw, "XDrawLine(%d,%d,%d,%d)\n", x, adjusted_y, x + width,
                  adjusted_y));
#endif
        XDrawLine(display, win, gc, x, adjusted_y, x + width, adjusted_y);
#endif
#ifdef DRAW_RESIZE_SHADOW
        DEBUGOUT(_XbaeDebug
                 (__FILE__, (Widget) mw, "DrawShadow(%d,%d,%d,%d)\n", shadow_width, x, adjusted_y,
                  width, height, XmSHADOW_OUT));
        DRAW_SHADOW(display, win, "win 7", mw->matrix.resize_top_shadow_gc,
                    mw->matrix.resize_bottom_shadow_gc, shadow_width, x, adjusted_y, width, height,
                    XmSHADOW_OUT);
#endif
        /*
         * Now draw the line (or shadow) on the non clipped region - that is
         * the fixed and trailingFixed columns.  First, do the leading columns.
         */
        if (mw->matrix.fixed_columns) {
                /*Dimension height = SOME_ROW_HEIGHT(mw, row);
                   local overwirting of predefined variable height... why??? */

                x = COLUMN_LABEL_OFFSET(mw);
                width = FIXED_COLUMN_WIDTH(mw) + 2 * mw->matrix.cell_shadow_thickness;
                win = XtWindow(mw);
                xbaeSetClipMask(mw, CLIP_FIXED_COLUMNS);

#ifdef DRAW_RESIZE_LINE
                if (XtIsManaged(LeftClip(mw))) {
#ifdef	VERBOSE_SLIDE
                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw, "XDrawLine(%d,%d,%d,%d)\n",
                                  -mw->matrix.cell_shadow_thickness, adjusted_y, width,
                                  adjusted_y));
#endif
                        XDrawLine(display, XtWindow(LeftClip(mw)), gc,
                                  -mw->matrix.cell_shadow_thickness, adjusted_y, width, adjusted_y);
                }
#endif
#ifdef DRAW_RESIZE_SHADOW
                if (XtIsManaged(LeftClip(mw)))
                        DRAW_SHADOW(display, XtWindow(LeftClip(mw)), "LeftClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width,
                                    -mw->matrix.cell_shadow_thickness, adjusted_y, width, height,
                                    XmSHADOW_OUT);
#endif
                xbaeSetClipMask(mw, CLIP_NONE);
        }

        /*
         * The trailingFixedRows
         */
        if (mw->matrix.trailing_fixed_columns) {
                x = TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw);
                width = TRAILING_FIXED_COLUMN_WIDTH(mw) + 2 * mw->matrix.cell_shadow_thickness;

                clip_reason = CLIP_TRAILING_FIXED_COLUMNS;
                if (IS_LEADING_FIXED_ROW(mw, row))
                        clip_reason |= CLIP_FIXED_ROWS;
                else if (IS_TRAILING_FIXED_ROW(mw, row))
                        clip_reason |= CLIP_TRAILING_FIXED_ROWS;

                xbaeSetClipMask(mw, clip_reason);

#ifdef DRAW_RESIZE_LINE
                if (XtIsManaged(RightClip(mw)))
                        XDrawLine(display, XtWindow(RightClip(mw)), gc,
                                  -mw->matrix.cell_shadow_thickness, adjusted_y, width, adjusted_y);
#endif
#ifdef DRAW_RESIZE_SHADOW
                if (XtIsManaged(RightClip(mw)))
                        DRAW_SHADOW(display, XtWindow(RightClip(mw)), "RightClip",
                                    mw->matrix.resize_top_shadow_gc,
                                    mw->matrix.resize_bottom_shadow_gc, shadow_width,
                                    -mw->matrix.cell_shadow_thickness, adjusted_y, width, height,
                                    XmSHADOW_OUT);
#endif
                xbaeSetClipMask(mw, CLIP_NONE);
        }

        if ((NEED_HORIZ_FILL(mw) && (!HAS_ATTACHED_TRAILING_COLUMNS(mw)))
            || need_horiz_dead_space_fill) {
                if (need_horiz_dead_space_fill) {
                        x = UNATTACHED_TRAILING_COLUMNS_OFFSET(mw) -
                            mw->matrix.cell_shadow_thickness;
                        width = 2 * mw->matrix.cell_shadow_thickness + HORIZ_DEAD_SPACE_WIDTH(mw);
                } else {
                        x = TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw) +
                            TRAILING_FIXED_COLUMN_WIDTH(mw);
                        width = FILL_HORIZ_WIDTH(mw) - VERT_SB_SPACE(mw);
                }

#ifdef DRAW_RESIZE_LINE
                XDrawLine(display, XtWindow(mw), gc, x, adjusted_y, width, adjusted_y);
#endif
#ifdef DRAW_RESIZE_SHADOW
                DRAW_SHADOW(display, XtWindow(mw), "win 8", mw->matrix.resize_top_shadow_gc,
                            mw->matrix.resize_bottom_shadow_gc, shadow_width, x, y, width, height,
                            XmSHADOW_OUT);
#endif
        }
}

/* Derived from SlideColumn */
static void SlideRow(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
        XbaeMatrixResizeColumnStruct *rd = (XbaeMatrixResizeColumnStruct *) data;
        XMotionEvent *motionEvent;
        Boolean relayout = False;
        int numCharacters;
        int i;
        /*int VERTICAL_RESIZE_HEIGHT = FONT_HEIGHT(rd->mw); */
        /*using 1 pixel resize step until further notice */
        int VERTICAL_RESIZE_HEIGHT = 1;
        /* SGO: implemented a limit so that rows can't be smaller than 5
           that does't makes sense in most cases. */
        int ROW_MINIMUM_HEIGHT = 5;

        if (event->type == ButtonRelease) {
                DEBUGOUT(_XbaeDebug(__FILE__, w, "SlideRow ButtonRelease()\n"));

                DrawSlideRow(rd->mw, rd->lasty);
                XUngrabPointer(XtDisplay(w), CurrentTime);
                rd->grabbed = False;
                /*
                 * Remanage the HSB if we unmapped it earlier
                 */
                if (rd->haveHSB)
                        XtManageChild(HorizScrollChild(rd->mw));

                if (rd->mw->matrix.resize_row_callback) {
                        XbaeMatrixResizeRowCallbackStruct call_data;

                        call_data.reason = XbaeResizeRowReason;
                        call_data.event = event;
                        call_data.row = rd->row - 1;
                        call_data.column = rd->column;
                        call_data.which = rd->row - 1;
                        call_data.rows = rd->mw->matrix.rows;
                        call_data.row_heights = rd->rowHeights;
                        XtCallCallbackList((Widget) rd->mw, rd->mw->matrix.resize_row_callback,
                                           (XtPointer) & call_data);
                }

                for (i = 0; i < rd->mw->matrix.rows; i++)
                        if (rd->rowHeights[i] != rd->mw->matrix.row_heights[i]) {
                                /* Make sure everything is handled correctly with SetValues */
                                XtVaSetValues((Widget) rd->mw, XmNrowHeights, rd->rowHeights, NULL);
                                break;
                        }
                XtFree((char *) rd->rowHeights);
                return;
        }

        if (event->type != MotionNotify)        /* Double check! */
                return;

#ifdef XBAE_VERY_VERBOSE
        DEBUGOUT(_XbaeDebug(__FILE__, (Widget) w, "SlideRow MotionNotify()\n"));
#endif
        motionEvent = (XMotionEvent *) event;

        if (rd->currenty - motionEvent->y > VERTICAL_RESIZE_HEIGHT) {
                /* If we're only one character wide, we cannae get any smaller */
                if (rd->rowHeights[rd->row - 1] == 0)
                        return;
                /*
                 * Moved up a step - update the column widths and force
                 * a redisplay
                 */
                numCharacters = (rd->currenty - motionEvent->y);
                if (numCharacters < 0)
                        numCharacters = 0;

                rd->rowHeights[rd->row - 1] -= numCharacters;
                rd->currenty -= numCharacters;

                if (rd->rowHeights[rd->row - 1] < ROW_MINIMUM_HEIGHT) {
                        rd->rowHeights[rd->row - 1] += numCharacters;
                        rd->currenty += numCharacters;
                        return;
                }
                relayout = True;
        }

        if (motionEvent->y - rd->currenty > VERTICAL_RESIZE_HEIGHT) {
                /*
                 * Moved down a step - update the column widths and force
                 * a redisplay
                 */
                numCharacters = (motionEvent->y - rd->currenty);
                rd->rowHeights[rd->row - 1] += numCharacters;
                rd->currenty += numCharacters;
                relayout = True;
        }

        if (relayout) {
                /* Draw the marker line in the new location */
                if (rd->lasty != rd->currenty) {
                        DrawSlideRow(rd->mw, rd->currenty);
                        DrawSlideRow(rd->mw, rd->lasty);

                        rd->lasty = rd->currenty;
                }
        }
}

/*
 * This is not a real action handler, it is called through xbaeResizeColumnsACT.
 * Therefore it can be declared static. (Real action handlers are referenced in
 * src/Matrix.c).
 */

/* ARGSUSED */
static void xbaeResizeRowsACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        int x, y, i, row, column;
        XbaeMatrixResizeColumnStruct resizeData;
        XtAppContext appcontext;
#ifdef DRAW_RESIZE_LINE
        XGCValues save, values;
        unsigned long gcmask;
#endif
        unsigned long event_mask;
        Display *display = XtDisplay(w);
        int fuzzy = XBAE_RESIZE_FUZZ;

        DEBUGOUT(_XbaeDebug(__FILE__, w, "xbaeResizeRowsACT()\n"));

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass)) {
                mw = (XbaeMatrixWidget) w;
        } else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass)) {
                mw = (XbaeMatrixWidget) XtParent(w);
        } else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "resizeColumnsACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to ResizeColumns action", NULL, 0);
                return;
        }

        /*
         * If we won't allow dynamic row resize, leave.
         */
        if (!mw->matrix.allow_row_resize)
                return;

        if (!xbaeEventToMatrixXY(mw, event, &x, &y))
                return;

        if (!xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column)) {
                DEBUGOUT(_XbaeDebug
                         (__FILE__, w, "xbaeResizeRowsAct return (cvt %d,%d to rowcol)\n", x, y));
                return;
        }
        DEBUGOUT(_XbaeDebug
                 (__FILE__, w, "xbaeResizeRowsAct-1: row %d y %d rh %d\n", row, y,
                  SOME_ROW_HEIGHT(mw, row)));

        /*
         * Figure out if we're resizing a row height...
         */
        if ((int) mw->matrix.cell_shadow_thickness > fuzzy)
                fuzzy = mw->matrix.cell_shadow_thickness;

        DEBUGOUT(_XbaeDebug
                 (__FILE__, w, "xbaeResizeRowsAct: row %d y %d fuz %d rh %d\n", row, y, fuzzy,
                  SOME_ROW_HEIGHT(mw, row)));
        if (y > fuzzy && SOME_ROW_HEIGHT(mw, row) - y > fuzzy)
                return;

        /*
         * Make sure we resize the right row. If y was just more than 0,
         * row is right, otherwise (if it's almost the old row height),
         * then increment row by one.
         */
        if (SOME_ROW_HEIGHT(mw, row) - fuzzy < y) {
                row++;
        }

        DEBUGOUT(_XbaeDebug(__FILE__, w, "xbaeResizeRowsACT() : resizing row %d\n", row));

        /*
         * Make it here and it's time to start the fun stuff!
         */
        /* Create the up/down cursor */
        if (mw->matrix.cursor)
                XFreeCursor(display, mw->matrix.cursor);
        mw->matrix.cursor = XCreateFontCursor(display, XC_sb_v_double_arrow);

        /* Commit any edit in progress and unmap the text field -
           it's just bad luck */
        (*((XbaeMatrixWidgetClass) XtClass(mw))->matrix_class.commit_edit) (mw, event, True);

        /*
         * Redraw the cell that had the text field in it or it might stay blank
         */
        xbaeDrawCell(mw, mw->matrix.current_row, mw->matrix.current_column);

        /*
         * Say goodbye to the Horizontal ScrollBar -> it only gets in the way!
         */
        if ((resizeData.haveHSB = XtIsManaged(HorizScrollChild(mw))
             && ((mw->matrix.scrollbar_placement == XmTOP_RIGHT)
                 || (mw->matrix.scrollbar_placement == XmBOTTOM_RIGHT))))
                XtUnmanageChild(HorizScrollChild(mw));
        /*
         * Flush the commit events out to the server.  Otherwise, our changes
         * to the GCs below have a bad effect.
         */
        XSync(display, False);

        event_mask = PointerMotionMask | ButtonReleaseMask;
        XtAddEventHandler(w, event_mask, True, (XtEventHandler) SlideRow, (XtPointer) & resizeData);

        XGrabPointer(display, XtWindow(w), True, event_mask, GrabModeAsync, GrabModeAsync,
                     XtWindow((Widget) mw), mw->matrix.cursor, CurrentTime);

        /* Copy the rowHeights array */
        resizeData.rowHeights = (short *) XtMalloc(mw->matrix.rows * sizeof(short));
        for (i = 0; i < mw->matrix.rows; i++)
                resizeData.rowHeights[i] = mw->matrix.row_heights[i];
        resizeData.grabbed = True;
        resizeData.mw = mw;
        resizeData.row = row;
        resizeData.starty = resizeData.currenty = resizeData.lasty = event->xbutton.y;

#ifdef DRAW_RESIZE_LINE
        gcmask = GCForeground | GCBackground | GCFunction;
        values.function = GXxor;
        XGetGCValues(display, mw->matrix.draw_gc, gcmask, &save);
        values.foreground = values.background = save.background;

        XChangeGC(display, mw->matrix.draw_gc, gcmask, &values);
#endif

        DrawSlideRow(mw, resizeData.currenty);

        appcontext = XtWidgetToApplicationContext(w);

        while (resizeData.grabbed)
                XtAppProcessEvent(appcontext, XtIMAll);

        XtRemoveEventHandler(w, event_mask, True, (XtEventHandler) SlideRow,
                             (XtPointer) & resizeData);

#ifdef DRAW_RESIZE_LINE
        XSetFunction(display, mw->matrix.draw_gc, GXcopy);
#endif
}

/*
 * This cannot be static - it is an action handler referenced in src/Matrix.c
 */

/* ARGSUSED */
void xbaeResizeColumnsACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        int x, y;
        int i;
        int row, column;
        XbaeMatrixResizeColumnStruct resizeData;
        XtAppContext appcontext;
#ifdef DRAW_RESIZE_LINE
        XGCValues save, values;
        unsigned long gcmask;
#endif
        unsigned long event_mask;
        Display *display = XtDisplay(w);
        int fuzzy = XBAE_RESIZE_FUZZ;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "resizeColumnsACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to ResizeColumns action", NULL, 0);
                return;
        }

        if (!xbaeEventToMatrixXY(mw, event, &x, &y)) {
                DEBUGOUT(_XbaeDebug
                         (__FILE__, (Widget) mw, "xbaeResizeColumnsACT: cannot convert to x/y\n"));
                return;
        }
        DEBUGOUT(_XbaeDebug(__FILE__, (Widget) mw, "xbaeResizeColumnsACT: x %d y %d\n", x, y));

        if (!xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column)) {
                DEBUGOUT(_XbaeDebug
                         (__FILE__, (Widget) mw,
                          "xbaeResizeColumnsACT: cannot convert to row/col\n"));
                return;
        }
        DEBUGOUT(_XbaeDebug
                 (__FILE__, (Widget) mw, "xbaeResizeColumnsACT: x %d y %d, row %d col %d\n", x, y,
                  row, column));

        /*
         * Figure out if we're resizing a row height...
         */
        if ((int) mw->matrix.cell_shadow_thickness > fuzzy)
                fuzzy = mw->matrix.cell_shadow_thickness;
        DEBUGOUT(_XbaeDebug
                 (__FILE__, (Widget) mw, "BETWEEN %d %d fuz %d allow %d\n", y,
                  SOME_ROW_HEIGHT(mw, row), fuzzy, mw->matrix.allow_row_resize));
        if (y <= fuzzy || SOME_ROW_HEIGHT(mw, row) - y <= fuzzy) {
                if (mw->matrix.allow_row_resize)
                        xbaeResizeRowsACT(w, event, params, nparams);
                return;
        }

        /*
         * If we won't allow dynamic column resize, leave.
         */
        if (!mw->matrix.allow_column_resize)
                return;

        /*
         * Calculate if the x and y of the middle button event is on
         * a column border.  Allow the width of the shadow to be the
         * allowed delta.  x is modified in xbaeXYToRowCol() to be
         * the x distance from the cell's border
         */
        if ((int) mw->matrix.cell_shadow_thickness > fuzzy)
                fuzzy = mw->matrix.cell_shadow_thickness;

        if (x > fuzzy && COLUMN_WIDTH(mw, column) - x > fuzzy)
                return;

        /*
         * Looks like we hit a column border, determine the column that is
         * intended to be resized
         */
        if ((COLUMN_WIDTH(mw, column) - x) <= fuzzy)
                column++;

        /* Can't adjust the origin or should you be able to?? */
        if (column == 0)
                return;

        /*
         * Make it here and it's time to start the fun stuff!
         */

        /* Create the left / right cursor */
        if (mw->matrix.cursor)
                XFreeCursor(display, mw->matrix.cursor);
        mw->matrix.cursor = XCreateFontCursor(display, XC_sb_h_double_arrow);

        /* Commit any edit in progress and unmap the text field -
           it's just bad luck */
        (*((XbaeMatrixWidgetClass) XtClass(mw))->matrix_class.commit_edit) (mw, event, True);

        /*
         * Redraw the cell that had the text field in it or it might stay blank
         */
        xbaeDrawCell(mw, mw->matrix.current_row, mw->matrix.current_column);

        /*
         * Say goodbye to the Vertical ScrollBar -> it only gets in the way!
         */
        if ((resizeData.haveVSB = XtIsManaged(VertScrollChild(mw))
             && ((mw->matrix.scrollbar_placement == XmTOP_RIGHT)
                 || (mw->matrix.scrollbar_placement == XmBOTTOM_RIGHT))))
                XtUnmanageChild(VertScrollChild(mw));
        /*
         * Flush the commit events out to the server.  Otherwise, our changes
         * to the GCs below have a bad effect.
         */
        XSync(display, False);

        event_mask = PointerMotionMask | ButtonReleaseMask;
        XtAddEventHandler(w, event_mask, True, (XtEventHandler) SlideColumn,
                          (XtPointer) & resizeData);

        XGrabPointer(display, XtWindow(w), True, event_mask, GrabModeAsync, GrabModeAsync,
                     XtWindow((Widget) mw), mw->matrix.cursor, CurrentTime);

        /* Copy the columnWidth array */
        resizeData.columnWidths = (short *) XtMalloc(mw->matrix.columns * sizeof(short));
        for (i = 0; i < mw->matrix.columns; i++)
                resizeData.columnWidths[i] = mw->matrix.column_widths[i];
        resizeData.grabbed = True;
        resizeData.mw = mw;
        resizeData.column = column;
        resizeData.startx = resizeData.currentx = resizeData.lastx = event->xbutton.x;

#ifdef DRAW_RESIZE_LINE
        gcmask = GCForeground | GCBackground | GCFunction;
        values.function = GXxor;
        XGetGCValues(display, mw->matrix.draw_gc, gcmask, &save);
        values.foreground = values.background = save.background;

        XChangeGC(display, mw->matrix.draw_gc, gcmask, &values);
#endif

        DrawSlideColumn(mw, resizeData.currentx);

        appcontext = XtWidgetToApplicationContext(w);

        while (resizeData.grabbed)
                XtAppProcessEvent(appcontext, XtIMAll);

        XtRemoveEventHandler(w, event_mask, True, (XtEventHandler) SlideColumn,
                             (XtPointer) & resizeData);

#ifdef DRAW_RESIZE_LINE
        XSetFunction(display, mw->matrix.draw_gc, GXcopy);
#endif
}

/*
 * Action to process a drag out
 */
void xbaeProcessDragACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        int x, y;
        int row, column;
        XbaeMatrixProcessDragCallbackStruct call_data;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "processDragACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to ProcessDrag action",
                                NULL, 0);
                return;
        }

        if (!mw->matrix.process_drag_callback)
                return;

        if (!xbaeEventToMatrixXY(mw, event, &x, &y))
                return;

        if (!xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column))
                return;

        call_data.reason = XbaeProcessDragReason;
        call_data.event = event;
        call_data.row = row;
        call_data.column = column;

        if (mw->matrix.draw_cell_callback) {
                Pixel bgcolor, fgcolor;
                int width, height, depth;

                call_data.type =
                    xbaeGetDrawCellValue(mw, row, column, &call_data.string, &call_data.pixmap,
                                         &call_data.mask, &width, &height, &bgcolor, &fgcolor,
                                         &depth);
        } else
                call_data.string = xbaeGetCellStringValue(mw, row, column);

        call_data.num_params = *nparams;
        call_data.params = params;

        XtCallCallbackList((Widget) mw, mw->matrix.process_drag_callback, (XtPointer) & call_data);
}

/*
 * Action to edit a non-fixed cell.
 */
void xbaeEditCellACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        int row, column;
        XrmQuark q;
        static XrmQuark QPointer, QLeft, QRight, QUp, QDown;
        static Boolean haveQuarks = False;
        /*
         * Get static quarks for the parms we understand
         */
        if (!haveQuarks) {
                QPointer = XrmPermStringToQuark("Pointer");
                QLeft = XrmPermStringToQuark("Left");
                QRight = XrmPermStringToQuark("Right");
                QUp = XrmPermStringToQuark("Up");
                QDown = XrmPermStringToQuark("Down");
                haveQuarks = True;
        }

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "editCellACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to EditCell action",
                                NULL, 0);
                return;
        }

        /*
         * Make sure we have a single parm
         */
        if (*nparams != 1) {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "editCellACT", "badParms",
                                "XbaeMatrix",
                                "XbaeMatrix: Wrong params passed to EditCell action, needs 1", NULL,
                                0);
                return;
        }

        /*
         * Initialize row/column to the current position
         */
        row = mw->matrix.current_row;
        column = mw->matrix.current_column;

        /*
         * Quarkify the string param
         */
        q = XrmStringToQuark(params[0]);

        /*
         * If we aren't currently editing, then the only kind of traversal that
         * makes sense is pointer.
         */
        if (!XtIsManaged(TextChild(mw)) && q != QPointer)
                return;

        if (q == QPointer) {
                int x, y;
                if (!xbaeEventToMatrixXY(mw, event, &x, &y)
                    || !xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column)
                    || (IS_FIXED(mw, row, column) && !mw->matrix.traverse_fixed)) {
                        return;
                }
        } else if (q == QRight) {
                /*
                 * If we are in the lower right corner, stay there.
                 * Otherwise move over a column. If we move off to the right of the
                 * final column to which traversing is allowed then move down a row
                 * and back to the first column to which traversing is allowed.
                 */
                if (!mw->matrix.traverse_fixed) {
                        /* check scrollable boundary */
                        if (mw->matrix.current_row != TRAILING_VERT_ORIGIN(mw) - 1
                            || mw->matrix.current_column != TRAILING_HORIZ_ORIGIN(mw) - 1) {
                                column++;
                                if (IS_TRAILING_FIXED_COLUMN(mw, column)) {
                                        column = mw->matrix.fixed_columns;
                                        row++;
                                }
                        }
                } else {
                        /* check matrix boundary */
                        if (mw->matrix.current_row != mw->matrix.rows - 1
                            || mw->matrix.current_column != mw->matrix.columns - 1) {
                                column++;
                                if (column >= mw->matrix.columns) {
                                        column = 0;
                                        row++;
                                }
                        }
                }
        } else if (q == QLeft) {
                /*
                 * If we are in the upper left corner, stay there.
                 * Otherwise move back a column. If we move before the first column
                 * to which traversing is allowed, move up a row and over to the last
                 * column to which traversing is allowed.
                 */
                if (!mw->matrix.traverse_fixed) {
                        /* check scrollable boundary */
                        if (mw->matrix.current_row != mw->matrix.fixed_rows
                            || mw->matrix.current_column != mw->matrix.fixed_columns) {
                                column--;
                                if (IS_LEADING_FIXED_COLUMN(mw, column)) {
                                        column = TRAILING_HORIZ_ORIGIN(mw) - 1;
                                        row--;
                                }
                        }
                } else {
                        if (mw->matrix.current_row != 0 || mw->matrix.current_column != 0) {
                                column--;
                                if (column < 0) {
                                        column = mw->matrix.columns - 1;
                                        row--;
                                }
                        }
                }
        } else if (q == QDown) {
                row++;

                /* adjust row for allowable traversable regions */
                if (!mw->matrix.traverse_fixed) {
                        if (IS_TRAILING_FIXED_ROW(mw, row))
                                row = mw->matrix.fixed_rows;
                } else {
                        if (row >= mw->matrix.rows)
                                row = 0;
                }
        } else if (q == QUp) {
                row--;

                if (!mw->matrix.traverse_fixed) {
                        if (IS_LEADING_FIXED_ROW(mw, row))
                                row = TRAILING_VERT_ORIGIN(mw) - 1;
                } else {
                        if (row < 0)
                                row = mw->matrix.rows - 1;
                }
        }

        /*
         * Call the traverseCellCallback to allow the application to
         * perform custom traversal.
         */
        if (mw->matrix.traverse_cell_callback) {
                XbaeMatrixTraverseCellCallbackStruct call_data;

                call_data.reason = XbaeTraverseCellReason;
                call_data.event = event;
                call_data.row = mw->matrix.current_row;
                call_data.column = mw->matrix.current_column;
                call_data.next_row = row;
                call_data.next_column = column;
                call_data.fixed_rows = mw->matrix.fixed_rows;
                call_data.fixed_columns = mw->matrix.fixed_columns;
                call_data.trailing_fixed_rows = mw->matrix.trailing_fixed_rows;
                call_data.trailing_fixed_columns = mw->matrix.trailing_fixed_columns;
                call_data.num_rows = mw->matrix.rows;
                call_data.num_columns = mw->matrix.columns;
                call_data.param = params[0];
                call_data.qparam = q;

                XtCallCallbackList((Widget) mw, mw->matrix.traverse_cell_callback,
                                   (XtPointer) & call_data);

                row = call_data.next_row;
                column = call_data.next_column;
        }

        /*
         * Attempt to edit the new cell using the edit_cell method.
         * If we are editing a cell based on pointer position, we always
         * call edit_cell.  Otherwise, we must be editing a new cell to
         * call edit_cell.
         */
        if (q == QPointer || (row != mw->matrix.current_row || column != mw->matrix.current_column))
                (*((XbaeMatrixWidgetClass) XtClass(mw))->matrix_class.edit_cell) (mw, event, row,
                                                                                  column, params,
                                                                                  *nparams);

        /*
         * Traverse to the textField
         */
        (void) XmProcessTraversal(TextChild(mw), XmTRAVERSE_CURRENT);
}

/*
 * Action to unmap the textField and discard any edits made
 */

/* ARGSUSED */
void xbaeCancelEditACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        Boolean unmap;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "cancelEditACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to CancelEdit action",
                                NULL, 0);
                return;
        }

        /*
         * Make sure we have a single param
         */
        if (*nparams != 1) {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "cancelEditACT", "badParms",
                                "XbaeMatrix",
                                "XbaeMatrix: Wrong params passed to CancelEdit action, needs 1",
                                NULL, 0);
                return;
        }

        /*
         * Validate our param
         */
        if (!strcmp(params[0], "True"))
                unmap = True;
        else if (!strcmp(params[0], "False"))
                unmap = False;
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "cancelEditACT", "badParm",
                                "XbaeMatrix", "XbaeMatrix: Bad parameter for CancelEdit action",
                                NULL, 0);
                return;
        }
        /*
         * Call the cancel_edit method
         */
        (*((XbaeMatrixWidgetClass) XtClass(mw))->matrix_class.cancel_edit) (mw, unmap);
}

/*
 * Action save any edits made and unmap the textField if params[0] is True
 */

/* ARGSUSED */
void xbaeCommitEditACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        Boolean unmap;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "commitEditACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to CommitEdit action",
                                NULL, 0);
                return;
        }

        /*
         * Make sure we have a single param
         */
        if (*nparams != 1) {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "commitEditACT", "badParms",
                                "XbaeMatrix",
                                "XbaeMatrix: Wrong params for CommitEdit action, needs 1", NULL, 0);
                return;
        }

        /*
         * Validate our param
         */
        if (!strcmp(params[0], "True"))
                unmap = True;
        else if (!strcmp(params[0], "False"))
                unmap = False;
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "commitEditACT", "badParm",
                                "XbaeMatrix", "XbaeMatrix: Bad parameter for CommitEdit action",
                                NULL, 0);
                return;
        }

        (void) (*((XbaeMatrixWidgetClass) XtClass(mw))->matrix_class.commit_edit) (mw, event,
                                                                                   unmap);
}

static int DoubleClick(XbaeMatrixWidget mw, XEvent * event, int row, int column)
{
        /* A double click in this instance is two clicks in the
           same cell in a time period < double_click_interval */
        Time current_time;
        unsigned long delta;
        static int ret = 0;
        static int lastButton = 0;

        if (event->type == ButtonRelease) {
                /* If the button is released, store the current location and time -
                   next time through, if it's a button press event, we check for
                   double click */
                mw->matrix.last_row = row;
                mw->matrix.last_column = column;
                if (ret)        /* just had a double click */
                        mw->matrix.last_click_time = (Time) 0;
                else
                        mw->matrix.last_click_time = event->xbutton.time;
                ret = 0;

                lastButton = event->xbutton.button;

                return ret;
        }

        current_time = event->xbutton.time;
        delta = current_time - mw->matrix.last_click_time;

        if (row == mw->matrix.last_row && column == mw->matrix.last_column
            && delta < (unsigned long) mw->matrix.double_click_interval)
                ret = 1;
        else
                ret = 0;

        if (event->xbutton.button != lastButton) {
                ret = 0;
        }
        return ret;
}

 /*ARGSUSED*/ static void PushButton(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
        XbaeMatrixButtonPressedStruct *button = (XbaeMatrixButtonPressedStruct *) data;
        XMotionEvent *motionEvent;
        int x, y;
        int row, column;
        Boolean pressed = button->pressed;

        DEBUGOUT(_XbaeDebug(__FILE__, w, "PushButton()\n"));

        if (event->type == ButtonRelease) {
                button->grabbed = False;
                XtRemoveGrab(w);

                if (button->pressed) {
                        /* If the button is still pressed, it has been released in the
                           same button that was pressed.  "Unpress" it and call the
                           callbacks */
                        if (button->column == -1)
                                xbaeDrawRowLabel(button->mw, button->row, False);
                        else if (button->row == -1)
                                xbaeDrawColumnLabel(button->mw, button->column, False);

                        if (button->mw->matrix.label_activate_callback) {
                                XbaeMatrixLabelActivateCallbackStruct call_data;

                                call_data.reason = XbaeLabelActivateReason;
                                call_data.event = event;
                                call_data.row_label = (button->column == -1);
                                call_data.row = button->row;
                                call_data.column = button->column;

                                if (button->column == -1)
                                        call_data.label =
                                            button->mw->matrix.row_labels[button->row];
                                else
                                        call_data.label =
                                            button->mw->matrix.column_labels[button->column];

                                XtCallCallbackList((Widget) button->mw,
                                                   button->mw->matrix.label_activate_callback,
                                                   (XtPointer) & call_data);
                        }
                }
                return;
        }

        if (event->type != MotionNotify)        /* We want to be sure about this! */
                return;

        motionEvent = (XMotionEvent *) event;
        x = motionEvent->x;
        y = motionEvent->y;

        if (!xbaeEventToMatrixXY(button->mw, event, &x, &y))
                return;

        if (xbaeMatrixXYToRowCol(button->mw, &x, &y, &row, &column))
                /* Moved off the labels */
                pressed = False;
        else {
                if (button->column != column || button->row != row)
                        /* Moved out of the button that was originally pressed */
                        pressed = False;
                else if (button->column == column || button->row == row)
                        pressed = True;
        }
        /* If the status of whether or not the button should be pressed has
           changed, redraw the appropriate visual */
        if (pressed != button->pressed) {
                if (button->column == -1)
                        xbaeDrawRowLabel(button->mw, button->row, pressed);
                else if (button->row == -1)
                        xbaeDrawColumnLabel(button->mw, button->column, pressed);
                /* And set our struct's pressed member to the current setting */
                button->pressed = pressed;
        }
}

 /*ARGSUSED*/ void xbaeHandleClick(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
        XbaeMatrixWidget mw = (XbaeMatrixWidget) data;
        int x, y;
        int row, column;
        Boolean in_rows, in_columns, in_cells, in_labels;

        DEBUGOUT(_XbaeDebug(__FILE__, w, "xbaeHandleClick()\n"));

        /* if we have a double click and a callback - break out! */
        if (event->type != ButtonPress && event->type != ButtonRelease)
                return;

        if (!xbaeEventToMatrixXY(mw, event, &x, &y)) {
                DEBUGOUT(_XbaeDebug(__FILE__, w, "xbaeEventToXY() fails\n"));
                return;
        }

        in_rows = xbaeMatrixYtoRow(mw, &y, &row);
        in_columns = xbaeMatrixXtoColumn(mw, &x, &column);
        in_cells = in_rows && in_columns && row != -1 && column != -1;
        in_labels = in_rows && in_columns && ((row == -1) ^ (column == -1));

        if (event->type == ButtonPress) {
                did_move = False;
                last_row = row;
                last_column = column;
        }

        if (in_labels
            && (mw->matrix.button_labels
                || (row == -1 && mw->matrix.column_button_labels
                    && mw->matrix.column_button_labels[column]) || (column == -1
                                                                    && mw->matrix.row_button_labels
                                                                    && mw->matrix.
                                                                    row_button_labels[row]))) {
                unsigned long event_mask;
                XtAppContext appcontext;
                XbaeMatrixButtonPressedStruct button;

                /*
                 * If it is ButtonRelease event, return - the ButtonRelease
                 * events are handled in the event handler loop below.
                 */
                if (event->type != ButtonPress) {
                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, w, "xbaeHandleClick - no ButtonPress, return\n"));
                        return;
                }

                if (column == -1) {
                        /* row label */
                        DEBUGOUT(_XbaeDebug(__FILE__, w, "Action in row label\n"));
                        xbaeDrawRowLabel(mw, row, True);
                } else if (row == -1) {
                        /* Column label */
                        DEBUGOUT(_XbaeDebug(__FILE__, w, "Action in column label\n"));
                        xbaeDrawColumnLabel(mw, column, True);
                }

                /* Action stations! */
                event_mask = ButtonReleaseMask | PointerMotionMask;

                XtAddGrab(w, True, False);
                /* Copy the data needed to be passed to the event handler */
                button.mw = mw;
                button.row = row;
                button.column = column;
                button.pressed = True;
                button.grabbed = True;

                XtAddEventHandler(w, event_mask, True, (XtEventHandler) PushButton,
                                  (XtPointer) & button);
                XtAddEventHandler(TextChild(mw), event_mask, True, (XtEventHandler) PushButton,
                                  (XtPointer) & button);

                appcontext = XtWidgetToApplicationContext(w);

                while (button.grabbed)
                        XtAppProcessEvent(appcontext, XtIMAll);

                XtRemoveEventHandler(w, event_mask, True, (XtEventHandler) PushButton,
                                     (XtPointer) & button);
                XtRemoveEventHandler(TextChild(mw), event_mask, True, (XtEventHandler) PushButton,
                                     (XtPointer) & button);

        }

        if (in_cells && mw->matrix.default_action_callback && w != (Widget) mw
            && DoubleClick(mw, event, mw->matrix.current_row, mw->matrix.current_column)) {
                XbaeMatrixDefaultActionCallbackStruct call_data;

                DEBUGOUT(_XbaeDebug(__FILE__, w, "xbaeHandleClick - DefaultActionCallback\n"));

                call_data.reason = XbaeDefaultActionReason;
                call_data.event = event;
                call_data.row = row;
                call_data.column = column;

                XtCallCallbackList((Widget) mw, mw->matrix.default_action_callback,
                                   (XtPointer) & call_data);
        } else {
                DEBUGOUT(_XbaeDebug
                         (__FILE__, w,
                          "xbaeHandleClick() uncaught case (x %d y %d, row %d col %d)\n", x, y, row,
                          column));
        }
}

/* ARGSUSED */
void xbaeTraverseNextACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w should be the textField widget.
         */
        if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "traverseNextACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to TraverseNext action", NULL, 0);
                return;
        }

        /*
         * Set the traversing direction flag.  XmProcessTraversal may traverse
         * to the Clip widget. If it does, then we will see this flag in
         * the Clip focusCallback, TraverseInCB, and we will continue to traverse
         * on out of the mw.  yuck!
         */
        mw->matrix.traversing = XmTRAVERSE_NEXT_TAB_GROUP;
        (void) XmProcessTraversal(TextChild(mw), XmTRAVERSE_NEXT_TAB_GROUP);
        mw->matrix.traversing = NOT_TRAVERSING;
}

/* ARGSUSED */
void xbaeTraversePrevACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w should be the textField widget.
         */
        if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "traversePrevACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to TraversePrev action", NULL, 0);
                return;
        }

        /*
         * Set the traversing direction flag.  XmProcessTraversal may traverse
         * to the Clip widget. If it does, then we will see this flag in
         * the Clip focusCallback, TraverseInCB, and we will continue to traverse
         * on out of the mw.  yuck!
         */
        mw->matrix.traversing = (int) XmTRAVERSE_PREV_TAB_GROUP;
        (void) XmProcessTraversal(TextChild(mw), XmTRAVERSE_PREV_TAB_GROUP);
        mw->matrix.traversing = NOT_TRAVERSING;
}

static void
callSelectCellCallbacks(XbaeMatrixWidget mw, XEvent * event, int row, int column, String * params,
                        Cardinal num_params)
{
        XbaeMatrixSelectCellCallbackStruct call_data;

        call_data.reason = XbaeSelectCellReason;
        call_data.event = event;
        call_data.row = row;
        call_data.column = column;
        call_data.num_params = num_params;
        call_data.params = params;
        call_data.cells = NULL; /* mw->matrix.cells */
        call_data.selected_cells = NULL;        /* mw->matrix.selected_cells */
        /*
         * Need to create a new structure and pass this to the callback.
         * Before the PER_CELL change, this was an easy action because
         * we could pass an existing structure. Now we need to build
         * it especially for this callback.
         */

        XtCallCallbackList((Widget) mw, mw->matrix.select_cell_callback, (XtPointer) & call_data);

}

/* ARGSUSED */
void xbaeSelectCellACT(Widget w, XEvent * event, String * params, Cardinal * num_params)
{
        XbaeMatrixWidget mw;
        int x, y;
        int row, column;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "xbaeSelectCellACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to SelectCell action",
                                NULL, 0);
                return;
        }

        /*
         * If we don't have a selectCellCallback, then return now
         */
        if (!mw->matrix.select_cell_callback)
                return;

        if (!xbaeEventToMatrixXY(mw, event, &x, &y))
                return;

        if (!xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column))
                return;

        /*
         * Call our select_cell callbacks
         */
        callSelectCellCallbacks(mw, event, row, column, params, *num_params);
}

 /*ARGSUSED*/ static void scrollSelect(XbaeMatrixScrollStruct * ss, int new_row, int new_column)
{
        if ((new_row != last_row || new_column != last_column)
            && (ss->mw->matrix.selection_policy == XmMULTIPLE_SELECT
                || ss->mw->matrix.selection_policy == XmEXTENDED_SELECT)) {
                Boolean old_scroll_select = ss->mw->matrix.scroll_select;
                ss->mw->matrix.scroll_select = False;

                if (!did_move) {
                        callSelectCellCallbacks(ss->mw, ss->event, last_row, last_column, NULL, 0);
                }

                String params[1] = { "extend" };
                callSelectCellCallbacks(ss->mw, ss->event, new_row, new_column, params,
                                        sizeof params / sizeof *params);

                ss->mw->matrix.scroll_select = old_scroll_select;

                did_move = True;
                last_row = new_row;
                last_column = new_column;
        }
}

static void checkScrollValues(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
        XbaeMatrixScrollStruct *ss = (XbaeMatrixScrollStruct *) data;
        Boolean in_rows, in_columns;
        int x, y;
        int row, column;

        if (event->type == ButtonRelease) {
                XtRemoveGrab(w);
                ss->grabbed = False;
        } else {
                if (!xbaeEventToMatrixXY(ss->mw, event, &x, &y))
                        return;

                ss->event = event;
                ss->currentx = x;
                ss->currenty = y;

                /* figure out the next row/column in which to select */

                in_rows = xbaeMatrixYtoRow(ss->mw, &y, &row);
                if (!in_rows || !(xbaeRowClip(ss->mw, row) & ss->scroll_region)) {
                        row = last_row;
                }

                in_columns = xbaeMatrixXtoColumn(ss->mw, &x, &column);
                if (!in_columns || !(xbaeColumnClip(ss->mw, column) & ss->scroll_region)) {
                        column = last_column;
                }

                scrollSelect(ss, row, column);
        }
}

static void updateScroll(XtPointer data)
{
        XbaeMatrixScrollStruct *ss = (XbaeMatrixScrollStruct *) data;
        int row = last_row;
        int column = last_column;

        if (!ss->grabbed)
                return;

        /*increase repeat count */
        ss->repeat_count++;

        /*
         * If we are off the clip widget and there is somthing that could
         * be scrolled into a visible position. Scroll them into view.  
         */

        /*Scroll now with aspect ratio (step 4 for vertical, step 3 for
           horizontal, for same speed on monitors with 4:3 aspect ratio) */

        if (ss->scroll_region & CLIP_VISIBLE_HEIGHT) {
                if (ss->currenty < FIXED_ROW_LABEL_OFFSET(ss->mw) && VERT_ORIGIN(ss->mw) > 0) {
                        /* above the non fixed cells */
                        xbaeScrollRows(ss->mw, True, Min(3, VERT_ORIGIN(ss->mw)));
                        row = xbaeTopRow(ss->mw);
                } else if (ss->currenty >= FIXED_ROW_LABEL_OFFSET(ss->mw) + VISIBLE_HEIGHT(ss->mw)
                           && VERT_ORIGIN(ss->mw) + VISIBLE_HEIGHT(ss->mw) <
                           NON_FIXED_TOTAL_HEIGHT(ss->mw)) {
                        /* bellow the non fixed cells */
                        xbaeScrollRows(ss->mw, False,
                                       Min(3,
                                           NON_FIXED_TOTAL_HEIGHT(ss->mw) - VERT_ORIGIN(ss->mw) -
                                           VISIBLE_HEIGHT(ss->mw)));
                        row = xbaeBottomRow(ss->mw);
                }
        }
        if (ss->scroll_region & CLIP_VISIBLE_WIDTH) {
                if (ss->currentx < FIXED_COLUMN_LABEL_OFFSET(ss->mw) && HORIZ_ORIGIN(ss->mw) > 0) {
                        xbaeScrollColumns(ss->mw, True, Min(4, HORIZ_ORIGIN(ss->mw)));
                        column = xbaeLeftColumn(ss->mw);
                } else if (ss->currentx >= FIXED_COLUMN_LABEL_OFFSET(ss->mw) + VISIBLE_WIDTH(ss->mw)
                           && HORIZ_ORIGIN(ss->mw) + VISIBLE_WIDTH(ss->mw) <
                           NON_FIXED_TOTAL_WIDTH(ss->mw)) {
                        xbaeScrollColumns(ss->mw, False,
                                          Min(4,
                                              NON_FIXED_TOTAL_WIDTH(ss->mw) - HORIZ_ORIGIN(ss->mw) -
                                              VISIBLE_WIDTH(ss->mw)));
                        column = xbaeRightColumn(ss->mw);
                }
        }

        scrollSelect(ss, row, column);

        /*
         * Flush the updates out to the server so we don't end up lagging
         * behind too far and end up with a million redraw requests.
         * Particularly for higher update speeds
         */

        /* adjust scrolling interval */
        if (ss->repeat_count < 6)
                ss->interval = DEFAULT_SCROLL_SPEED / ss->repeat_count;

        XFlush(XtDisplay((Widget) ss->mw));

        ss->timerID =
            XtAppAddTimeOut(ss->app_context, ss->interval, (XtTimerCallbackProc) updateScroll,
                            (XtPointer) ss);
}

/* ARGSUSED */
void xbaeHandleMotionACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        int x, y, row, column;
        XbaeMatrixScrollStruct scrollData;
        static Boolean scrolling = False;

        if (scrolling)
                return;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w could be Matrix, or the Clip or textField children of Matrix
         */
        if (XtIsSubclass(w, xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) w;
        else if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "handleMotionACT", "badWidget",
                                "XbaeMatrix",
                                "XbaeMatrix: Bad widget passed to HandleMotion action", NULL, 0);
                return;
        }

        if (!xbaeEventToMatrixXY(mw, event, &x, &y))
                return;

        /*
         * In this instance, we don't care if a valid row and column are
         * returned as we'll be the judge of the result
         */
        xbaeMatrixXYToRowCol(mw, &x, &y, &row, &column);

        /*
         * Grab the pointer and add a timeout routine to start modifying
         * the current row and/or column in the matrix.  Also add an
         * event handler to monitor the current distance outside the
         * matrix so we can adjust the timeout routine to go faster when
         * the pointer is further away from the matrix.
         */

        scrolling = True;

        XtAddGrab(w, True, False);

        scrollData.mw = mw;
        scrollData.event = event;
        scrollData.interval = DEFAULT_SCROLL_SPEED;
        scrollData.grabbed = True;
        scrollData.app_context = XtWidgetToApplicationContext(w);
        scrollData.repeat_count = 0;
        scrollData.scroll_region = xbaeCellClip(mw, last_row, last_column);

        XtAddEventHandler(w, PointerMotionMask | ButtonReleaseMask, True,
                          (XtEventHandler) checkScrollValues, (XtPointer) & scrollData);
        /*
         * Call checkScrollValues() to find out where exactly we are in
         * relation to the clip widget
         */
        checkScrollValues(w, (XtPointer) & scrollData, event, NULL);

        /*
         * The above / below / left / right members of the scrollData struct
         * should now be set so we know where we should be moving.  Let's
         * get on with it, eh?
         */
        updateScroll((XtPointer) & scrollData);

        while (scrollData.grabbed)
                XtAppProcessEvent(scrollData.app_context, XtIMAll);

        XtRemoveEventHandler(w, PointerMotionMask | ButtonReleaseMask, True,
                             (XtEventHandler) checkScrollValues, (XtPointer) & scrollData);

        /*
         * We don't want the timeout getting called again as, in two lines,
         * we'll be way out of scope!
         */
        XtRemoveTimeOut(scrollData.timerID);
        scrolling = False;
}

/* ARGSUSED */
void xbaePageDownACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        char *down = "0";
        int top;
        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w should be the textField widget.
         */
        if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "pageDownACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to PageDown action",
                                NULL, 0);
                return;
        }

        if (!XtIsManaged(VertScrollChild(mw)))
                return;

        /*
         * Save the top row - if scrolling occurs, the text widget needs
         * to be moved
         */
        top = TOP_ROW(mw);      /* SGO:TOP_ROW returns current first row in row numbers
                                 * VERT_ORIGIN returns current first row in pixel position
                                 */

        XtCallActionProc(VertScrollChild(mw), "PageDownOrRight", event, &down, 1);

        if (TOP_ROW(mw) != top)
                /*
                 * Position the cursor at the top most non fixed row if there was
                 * a page down
                 */
                XbaeMatrixEditCell((Widget) mw, TOP_ROW(mw) + mw->matrix.fixed_rows,
                                   mw->matrix.current_column);
}

/* ARGSUSED */
void xbaePageUpACT(Widget w, XEvent * event, String * params, Cardinal * nparams)
{
        XbaeMatrixWidget mw;
        char *up = "0";
        int top;

        /*
         * Get Matrix widget and make sure it is a Matrix subclass.
         * w should be the textField widget.
         */
        if (XtIsSubclass(XtParent(w), xbaeMatrixWidgetClass))
                mw = (XbaeMatrixWidget) XtParent(w);
        else {
                XtAppWarningMsg(XtWidgetToApplicationContext(w), "pageUpACT", "badWidget",
                                "XbaeMatrix", "XbaeMatrix: Bad widget passed to PageUp action",
                                NULL, 0);
                return;
        }

        if (!XtIsManaged(VertScrollChild(mw)))
                return;

        /*
         * Save the top row - if scrolling occurs, the text widget needs
         * to be moved
         */
        top = TOP_ROW(mw);

        XtCallActionProc(VertScrollChild(mw), "PageUpOrLeft", event, &up, 1);

        if (TOP_ROW(mw) != top)
                /*
                 * Position the cursor at the top most non fixed row if there was
                 * a page down
                 */
                XbaeMatrixEditCell((Widget) mw, TOP_ROW(mw) + mw->matrix.fixed_rows,
                                   mw->matrix.current_column);
}
